/*
 * Copyright (c) 2011, 2012, Endea
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

package endea.internal.entity

import endea._
import java.lang.reflect.Field
import java.lang.reflect.Modifier
import endea.internal._
import endea.io._
import java.io.BufferedWriter
import java.io.FileWriter
import java.security.SecureRandom
import scala.collection.mutable.ArrayBuffer
import scala.collection.mutable.ListBuffer
import scala.collection.mutable.HashMap
import java.util.concurrent.atomic.AtomicInteger

object MetaEntity {

  lazy val idField = {
    val id = classOf[Entity].getDeclaredField("id")
    id.setAccessible(true)
    id
  }

  lazy val all = {

    val entities = new ListBuffer[Class[_ <: Entity]]

    val classLoader = classOf[Module].getClassLoader()
    for (
      module <- Module.modules;
      entity <- module.entities
    ) {
      try {
        // TODO test class is assignable Class[Entity]
        entities += classLoader.loadClass(entity).asInstanceOf[Class[_ <: Entity]]
      } catch {
        case e: ClassNotFoundException =>
          System.err.println(entity + " not found - " + module)
      }
    }

    entities
  }

  /////

  private val idToMetaEntity = new HashMap[Int, MetaEntity]
  private val classToMetaEntity = new HashMap[String, MetaEntity]

  private[entity] val directory = Endea.path \\ "metadata"

  private val sequenceFile = directory \ "sequence"

  private val id = new AtomicInteger(
    if (sequenceFile.exists) sequenceFile.getLines().next().toInt
    else 1)

  def apply[E <: Entity](implicit manifest: Manifest[E]) = getByClass(manifest.erasure)

  def getByObject(any: AnyRef) = getByClass(any.getClass())

  // def getByClass[T](implicit manifest: Manifest[T]) = getByClass(manifest.erasure)

  def getByClass(clazz: Class[_]): MetaEntity = {

    val className = clazz.getName

    val option = classToMetaEntity.get(className)
    if (option.isDefined)
      option.get
    else {
      val link = toFile(clazz)
      if (link.exists()) {
        load(link)
      } else {
        val metaEntity = make(clazz)
        save(metaEntity)
        updateMap(metaEntity)
      }
    }
  }

  def getById(id: Int): Option[MetaEntity] = {

    val option = idToMetaEntity.get(id)
    if (option.isDefined)
      option
    else {
      val file = toFile(id)
      if (!file.exists())
        None
      else {
        Some(load(file))
      }
    }
  }

  private def updateMap(metaEntity: MetaEntity) = {
    idToMetaEntity.put(metaEntity.id, metaEntity)
    classToMetaEntity.put(metaEntity.clazz.getName, metaEntity)
    metaEntity
  }

  private[internal] def make(clazz: Class[_]): MetaEntity = {

    val attributes = for (
      field <- clazz.getDeclaredFields();
      if (!Modifier.isTransient(field.getModifiers()))
    ) yield new Attribute(nextId(), field)

    new MetaEntity(nextId(), clazz, attributes)
  }

  private[entity] def load(file: File): MetaEntity = {

    var id = -1
    var clazz: Class[_] = null
    val attributes = new ArrayBuffer[Attribute]

    var eof = false

    try {
      for (line <- file.getLines()) {
        if (!eof && line.length() > 0) {
          val tokens = line.trim.split("\\s+")
          if (tokens.size > 0) {
            val select = tokens(0)(0)
            if (select == 'E') {
              eof = true
            } else if (select == '[') {
              id = tokens(0).drop(1).toInt
              clazz = Class.forName(tokens(1))
            } else {
              attributes += new Attribute(tokens(0).toInt, clazz.getDeclaredField(tokens(1)))
            }
          }
        }
      }
    } catch {
      case e: Exception =>
        throw new Exception("Corrupted: " + file.name, e)
    }

    if (!eof) {
      throw new Exception("No EOF: " + file.name)
    }

    updateMap(new MetaEntity(id, clazz, attributes.toSeq))
  }

  private[internal] def save(meta: MetaEntity) {

    sequenceFile < id.get.toString

    val sb = new StringBuilder()
    sb += '['
    sb ++= meta.id.toString
    sb + ' '
    sb ++= meta.clazz.getName
    sb += '\n'

    for (attribute <- meta.declaredAttributes) {
      sb ++= attribute.id.toString
      sb += ' '
      sb ++= attribute.name
      sb += '\n'
    }
    sb ++= "EOF"
    val file = toFile(meta.id, 1)
    file < sb.toString

    toFile(meta.clazz).link(file)
    toFile(meta.id).link(file)

  }

  private[entity] def toFile(clazz: Class[_]) = directory \ clazz.getName

  private[entity] def toFile(id: Int, version: Int = 0) = directory \ (if (version == 0) id.toString else
    id.toString + "." + version.toString)

  private def nextId() = id.getAndIncrement()

  private[entity] def clear() {

    if (!Endea.test)
      throw new IllegalStateException("No test mode!")
    else {
      idToMetaEntity.clear()
      classToMetaEntity.clear()
      directory.clear()
      id.set(1)
    }
  }
}

class MetaEntity private (
  val id: Int,
  val clazz: Class[_],
  val declaredAttributes: Seq[Attribute]) {

  private var className: String = _

  lazy val attributes: Seq[Attribute] = {

    val parent = clazz.getSuperclass
    if (parent == classOf[Entity])
      declaredAttributes
    else
      MetaEntity.getByClass(parent).attributes ++ declaredAttributes
  }

  private lazy val instantiator = new org.objenesis.ObjenesisStd().getInstantiatorOf(clazz)

  def newInstance(): Entity = instantiator.newInstance().asInstanceOf[Entity]

  def attribute(id: Int): Option[Attribute] = {

    for (attribute <- attributes)
      if (attribute.id == id)
        return Some(attribute)

    None
  }

  def findAttribute(name: String): Option[Attribute] = {

    for (attribute <- attributes)
      if (attribute.name == name)
        return Some(attribute)

    None
  }

  override def toString() = {

    val sb = new StringBuilder
    sb ++= clazz.getSimpleName
    sb += '@'
    sb ++= id.toString
    sb += '['
    for (attribute <- attributes) {
      sb ++= "( "
      sb ++= attribute.id.toString
      sb ++= " -> "
      sb ++= attribute.name
      sb ++= ") "
    }
    sb += ']'

    sb.toString()
  }
}
