/*
 * Copyright (c) 2011, Endea.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

package endea.internal.entity

import endea._
import endea.io._
import endea.value._

import scala.collection.mutable.ListBuffer
import scala.collection.mutable.ArrayBuffer

object Serial {

  private val INT: Byte = 1
  private val LONG: Byte = 2
  private val FLOAT: Byte = 3
  private val DOUBLE: Byte = 4
  private val BOOLEAN: Byte = 5
  private val STRING: Byte = 6
  private val ENTITY: Byte = 7
  private val VALUE: Byte = 8
  private val SEQ: Byte = 9

  final def write(entity: Entity, out: ByteOutput) {

    val metaEntity = MetaEntity.getByObject(entity)

    out.putVarInt(metaEntity.id)
    out.putVarLong(entity.id)

    /////

    for (attribute <- metaEntity.attributes) {

      val value = attribute.get(entity)
      value match {
        case null =>
        case None =>
        case s: Seq[_] if s.isEmpty =>
        case value =>
          out.putVarInt(attribute.id)
          writeValue(out, value)
      }
    }

  }

  def read(input: ByteInput): Entity = {

    val entityClassId = input.getVarInt()
    val metaEntity = MetaEntity.getById(entityClassId).
      getOrElse(throw new RuntimeException("Entity class not found: " + entityClassId.toHexString))

    val entity = metaEntity.newInstance()

    MetaEntity.idField.set(entity, input.getVarLong())

    /////

    while (input.hasNext()) {

      val attributeId = input.getVarInt()

      var value = readValue(input)
      if (value != null) {
        val optAttribute = metaEntity.attribute(attributeId)
        if (optAttribute == null)
          throw new Exception(attributeId + "? " + MetaEntity)
        else {
          val attribute = optAttribute.get

          if (value.isInstanceOf[Seq[_]]) {
            var seq = value.asInstanceOf[Seq[_]]
            if (attribute.isValue) {
              val pType = attribute.field.getAnnotation(classOf[PType])
              if (pType == null)
                throw new Exception("Declare @PType to " + attribute.field)
              val constructor = getConstructor(pType.value)
              seq = seq.map(item => constructor.newInstance(item.asInstanceOf[AnyRef]))
            }

            val _type = attribute.getType
            if (classOf[List[_]].isAssignableFrom(_type)) seq = seq.toList
            else if (classOf[IndexedSeq[_]].isAssignableFrom(_type)) seq = seq.toIndexedSeq
            else throw new Exception(" seq.toX not implemented yet for : " + _type.getName) // TODO

            value = seq

          } else if (attribute.getType.isAssignableFrom(classOf[Some[_]]))
            value = Some(value)
          else if (attribute.isValue)
            value = getConstructor(attribute.getType).newInstance(value.asInstanceOf[AnyRef])

          attribute.set(entity, value)
        }
      }
    }

    for (attribute <- metaEntity.attributes) {
      attribute.get(entity) match {
        case null =>
          val _type = attribute.getType
          if (_type.isAssignableFrom(classOf[Option[_]])) {
            attribute.set(entity, None)
          } else if (_type.isAssignableFrom(classOf[List[_]])) {
            attribute.set(entity, Nil)
          } else if (_type.isAssignableFrom(classOf[IndexedSeq[_]])) {
            attribute.set(entity, IndexedSeq.empty)
          }
        case _ =>
      }
    }

    entity.asInstanceOf[Entity]
  }

  private def writeValue(out: ByteOutput, value: Any) {

    value match {
      case i: Int => out.put(INT); out.putInt(i)
      case l: Long => out.put(LONG); out.putLong(l)
      case f: Float => out.put(FLOAT); out.putFloat(f)
      case d: Double => out.put(DOUBLE); out.putDouble(d)
      case b: Boolean => out.put(BOOLEAN); out.putBoolean(b)
      case s: String => out.put(STRING); out.putString(s)

      case entity: Entity =>
        out.put(ENTITY)
        out.putVarLong(entity.id)

      case value: Value[_] =>
        writeValue(out, value.asValue)

      case list: Seq[_] =>
        out.put(SEQ)
        out.putVarInt(list.length)
        for (item <- list)
          writeValue(out, item)

      case Some(value) =>
        writeValue(out, value)

      case other => throw new Exception("Unsupported type: " + other)
    }

  }

  private def readValue(input: ByteInput): Any = {

    input.get() match {

      case INT => input.getInt().asInstanceOf[java.lang.Integer]
      case LONG => input.getLong().asInstanceOf[java.lang.Long]
      case FLOAT => input.getFloat().asInstanceOf[java.lang.Float]
      case DOUBLE => input.getDouble().asInstanceOf[java.lang.Double]
      case BOOLEAN => input.getBoolean().asInstanceOf[java.lang.Boolean]
      case STRING => input.getString()

      case ENTITY =>
        Entity.get(input.getVarLong()) match {
          case Some(entity) => entity
          case _ => // TODO deleted entity
        }

      case SEQ =>
        val seq = new ArrayBuffer[Any]()
        for (_ <- 1 to input.getVarInt())
          seq += readValue(input)
        seq

      case other => throw new Exception("Unkown type: " + other)
    }
  }

  def getConstructor(clazz: Class[_]) = {

    val pType = clazz.getAnnotation(classOf[PType])
    if (pType == null)
      throw new Exception("Declare @PType to " + clazz.getName)

    val option = clazz.getDeclaredConstructors.find(constructor => {
      val types = constructor.getParameterTypes()
      types.length == 1 && types(0) == pType.value
    })

    if (option.isDefined) {
      val constructor = option.get
      constructor.setAccessible(true)
      constructor
    } else
      throw new Exception("Constructor not found for: " + clazz.getName)
  }
}